#include "imesh.h"
#include <stdio.h>
#include "macro_codes.h"
#include <math.h>
#include "dmavif1.h"
#include "imeshinstance.h"
#include "stitchhack.h"


// TRI
const uint64 orgGIFTagStrip[2] = 
{ 
	GIFTAG(0, 0, 1, PRIM_SET(4, 1, 1, 0, 0, 0, 1, 0, 0), 0, 3), 
	GIFTAG_REGS4(3, 1, 4, 0)
};
   
const uint64 orgGIFTagTriangle[2] =
{ 
	GIFTAG(0, 0, 1, PRIM_SET(3, 1, 1, 0, 0, 0, 1, 0, 0), 0, 9), 
	GIFTAG_REGS12(3, 1, 4, 3, 1, 4, 3, 1, 4, 0, 0, 0)
};




IMeshInstance * IMesh::CreateInstance(CDmaVif1 * pInstDmaVif1)
{
	IMeshInstance * pMeshInst = NULL;

	// Has a mesh been built yet?
	if(m_iNumOfChunks <= 0)
		return NULL;

	if((pMeshInst=new IMeshInstance(this, pInstDmaVif1)) == NULL) 
		return NULL;
	else
		return pMeshInst;
}

char SnapToChar(int a)
{
	float fTmp;
	if(a<0) fTmp = a-0.5f;
	else fTmp = a+0.5f;

	if(fTmp>0x7f) fTmp = 0x7f;
	else if(fTmp < -0x7f) fTmp = -0x7f;

	return (char) fTmp;
}

int SnapScale(float fVal, float fScale)
{
	float fRes = fVal * fScale;
	if(fRes<0) fRes-=0.5f;
	else fRes+=0.5f;

	if(fRes<0 && (-fRes)>0x7fffffffL) fRes = -0x7fffffffL;
	else if(fRes>=0 && fRes>0x7fffffffL) fRes = 0x7fffffffL;

	return (int) fRes;
}

uint32 * SetRow(CDmaVif1 * pDmaVif1, uint32 * pPtr)
{
	AddWordToChain(pDmaVif1, STROW);
	AddWordToChain(pDmaVif1, *pPtr++);
	AddWordToChain(pDmaVif1, *pPtr++);
	AddWordToChain(pDmaVif1, *pPtr++);
	AddWordToChain(pDmaVif1, *pPtr++);
	return pPtr;
}

uint32 * PreLoad4Verts(CDmaVif1 * pDmaVif1, SStripChunk &chunk, uint32 uVertsPassed, uint32 uCurBuf, int * pStripStart, int * pNumSkipped)
{				
	// number of vertices in the first strip in the next chunk
	int iPreNumVsInStrip = chunk.m_pNumVertsInStrip[0];
	*pStripStart = (iPreNumVsInStrip <= 4) ? 1 : 0;

	if(iPreNumVsInStrip == 3)
		*pNumSkipped = 1;
	else if(iPreNumVsInStrip == 4)
		*pNumSkipped = 0;
	else
	{
		assert(iPreNumVsInStrip > 4);
		*pNumSkipped = 4;
	}
					
	uint32 * pVPtr = (uint32 *) chunk.m_pVerts;
	pVPtr = SetRow(pDmaVif1, pVPtr);
	int iWrites = (iPreNumVsInStrip > 4) ? 4 : iPreNumVsInStrip;
	AddWordToChain(pDmaVif1, UNPACK(V4_8, iWrites, PAD_CHUNK+2+3*uVertsPassed+uCurBuf+1) );
	for(int iN=0; iN<iWrites; iN++)
		AddWordToChain(pDmaVif1, *pVPtr++);
	if(iPreNumVsInStrip <= 4 && chunk.m_iNumOfStrips>1)
		pVPtr = SetRow(pDmaVif1, pVPtr);
	if(iPreNumVsInStrip < 4 && chunk.m_iNumOfStrips>1)
	{
		assert(iPreNumVsInStrip == 3);
		AddWordToChain(pDmaVif1, UNPACK(V4_8, 1, PAD_CHUNK+2+3*(3+uVertsPassed)+uCurBuf+1) );
		AddWordToChain(pDmaVif1, *pVPtr++);
	}

	return pVPtr;
}
				


bool IMesh::ConvertToChunk(CDmaVif1 * pDmaVif1, SRenderGadget * pAllStrips, int iTotalNumOfStrips)
{
	CleanUp();

	// Set chunkSize
	m_iNumOfChunks = 1;	// Initial value
	int iChunkNumQuads = HUNK_SIZE;
	int iStripNumInChunk = 0;

	for(int iC=0; iC<iTotalNumOfStrips; iC++)
	{
		int iTstAlloc = iChunkNumQuads - pAllStrips[iC].m_iNumElems*3;	

		if(iTstAlloc < 0)	// Didn't find space in chunk
		{
			++m_iNumOfChunks;	// move on to next chunk
			iChunkNumQuads = HUNK_SIZE - pAllStrips[iC].m_iNumElems*3;			
			iStripNumInChunk = 0;
		}
		else
			iChunkNumQuads = iTstAlloc;

		++iStripNumInChunk;
	}

	// Alloc Hunks
	m_pChunks = NULL;

	if((m_pChunks=new SStripChunk[m_iNumOfChunks])==NULL)
	{ CleanUp(); return false; }


	// Build Hunks
	int iStrip = 0;

	for(int iCHunk=0; iCHunk<m_iNumOfChunks; iCHunk++)
	{
		SStripChunk &Chnk = m_pChunks[iCHunk];
		iChunkNumQuads = HUNK_SIZE;
		Chnk.m_iNumOfStrips = 0;
		Chnk.m_pNumVertsInStrip = NULL;
		Chnk.m_pVerts = NULL;
		Chnk.m_pCols = NULL;
		Chnk.m_pTexCrds = NULL;
		//Chnk.m_uGifTagStrip = *((u_long128 *) &orgGIFTagStrip[0]);
		//Chnk.m_uGifTagTriangle = *((u_long128 *) &orgGIFTagTriangle[0]);
		((uint64 *) &Chnk.m_uGifTagStrip)[0] = orgGIFTagStrip[0];
		((uint64 *) &Chnk.m_uGifTagStrip)[1] = orgGIFTagStrip[1];

		((uint64 *) &Chnk.m_uGifTagTriangle)[0] = orgGIFTagTriangle[0];
		((uint64 *) &Chnk.m_uGifTagTriangle)[1] = orgGIFTagTriangle[1];


		while( iStrip<iTotalNumOfStrips && (iChunkNumQuads - pAllStrips[iStrip].m_iNumElems*3) >= 0)
		{
			++Chnk.m_iNumOfStrips;

			 iChunkNumQuads -= pAllStrips[iStrip].m_iNumElems*3;
			++iStrip;
		}

		int iSizeInQuads = Chnk.m_iNumOfStrips>>2;	// was zero
		if((Chnk.m_iNumOfStrips&0x3)!=0)
			++iSizeInQuads;

		if((Chnk.m_pNumVertsInStrip=(int *) (new uint128[iSizeInQuads])) == NULL)
		{ CleanUp(); return false; }
	}

	// Get Strips
	iStrip = 0;
	float fMaxDeltaSizeVert = 0;
	for(int iCHunk=0; iCHunk<m_iNumOfChunks; iCHunk++)
	{
		SStripChunk &Chnk = m_pChunks[iCHunk];
		Chnk.m_iNumElemsInChunk = 0;
		Chnk.m_pVerts = NULL; 
		Chnk.m_pCols = NULL;
		Chnk.m_pTexCrds = NULL;

		for(int iS=0; iS<Chnk.m_iNumOfStrips; iS++)
		{
			Chnk.m_iNumElemsInChunk += pAllStrips[iStrip+iS].m_iNumElems;
			Chnk.m_pNumVertsInStrip[iS] = pAllStrips[iStrip+iS].m_iNumElems;	// Was zero
		}

		if((Chnk.m_pVerts=new float[4*Chnk.m_iNumElemsInChunk]) == NULL)
		{ CleanUp(); return false; }

		if((Chnk.m_pCols=new int[4*Chnk.m_iNumElemsInChunk]) == NULL)
		{ CleanUp(); return false; }

		if((Chnk.m_pTexCrds=new float[4*Chnk.m_iNumElemsInChunk]) == NULL)
		{ CleanUp(); return false; }

		// Fill up with vertices
		int iElem = 0;
		for(int iS=0; iS<Chnk.m_iNumOfStrips; iS++)
		{
			SRenderGadget &sCurStrip = pAllStrips[iStrip+iS];
			

			for(int iCnt=0; iCnt<sCurStrip.m_iNumElems; iCnt++)
			{
				((uint128 *) Chnk.m_pVerts)[iElem] = ((uint128 *) sCurStrip.m_pVerts)[iCnt];
				((uint128 *) Chnk.m_pCols)[iElem] = ((uint128 *) sCurStrip.m_pCols)[iCnt];
				((uint128 *) Chnk.m_pTexCrds)[iElem] = ((uint128 *) sCurStrip.m_pTexCrds)[iCnt];

				if(iCnt > 0)
				{
					for(int iT=0; iT<3; iT++)
					{
						float fAbsDeltaVert = fabs(sCurStrip.m_pVerts[4*iCnt + iT] - sCurStrip.m_pVerts[4*(iCnt-1) + iT]);
						
						if(fAbsDeltaVert > fMaxDeltaSizeVert)
							fMaxDeltaSizeVert = fAbsDeltaVert;
					}
				}
				
				++iElem;
			}
		}

		iStrip += Chnk.m_iNumOfStrips;
	}
	
	// Pack strips to prepare for difference mode
	uint32 uADCTimer = 0;	// need to move ADC occurence 4 forward
	uint32 uADCTimerClose = 0;	// Close ADC after 2 vertices
	bool bADC = false;	// leave at off
	uint32 uRowWfield;	// current value of w in row reg
	
	for(int iCHunk=0; iCHunk<m_iNumOfChunks; iCHunk++)
	{
		SStripChunk &Chnk = m_pChunks[iCHunk];

		// locate memory for current chunk
		uint32 * pDstVerts = (uint32 *) Chnk.m_pVerts;
		uint32 * pDstSTs = (uint32 *) Chnk.m_pTexCrds;
		uint32 * pDstCols = (uint32 *) Chnk.m_pCols;

		float fVScale = 126.0f / fMaxDeltaSizeVert;
		float * pVSrcData = Chnk.m_pVerts;
		float * pSTSrcData = Chnk.m_pTexCrds;
		uint32 * pColSrcData = (uint32 *) Chnk.m_pCols;
		
		for(int iS=0; iS<Chnk.m_iNumOfStrips; iS++)
		{
			int iNumVertsInStrip = Chnk.m_pNumVertsInStrip[iS];

			// necessary to backup 5 vals to use same buffer for dstmem
			float fRecVx = pVSrcData[0], fRecVy = pVSrcData[1];
			float fRecVz = pVSrcData[2], fRecVw = pVSrcData[3];
			float fRecVx2 = pVSrcData[4];

			// first four Vert longs for ROW register
			*pDstVerts++ = SnapScale(pVSrcData[0], fVScale);
			*pDstVerts++ = SnapScale(pVSrcData[1], fVScale);
			*pDstVerts++ = SnapScale(pVSrcData[2], fVScale);
			*pDstVerts++ = 0x7fffL;		// Set ADC bit
			uRowWfield = 0x7fffL;


			char cX=0, cY=0, cZ=0;

			uADCTimer |= 0x10;
			
			for(int i=0; i<iNumVertsInStrip; i++)
			{
				// Tst for ADC on
				
				if((uADCTimer&1)==1)
				{
					bADC = true;	// set to 0x8000
					assert(uADCTimerClose == 0);
					uADCTimerClose = 0x4;
				}
				else if((uADCTimerClose&1)==1)
				{
					bADC = false;
				}

				char adc = 0;
				if(bADC == true && uRowWfield == 0x7fff)
				{ adc = 1; uRowWfield = 0x8000; }
				else if(bADC == false && uRowWfield == 0x8000)
				{ adc = -1; uRowWfield = 0x7fff; } 
				
				// Fill up current vertex
				((char *) pDstVerts)[0] = cX;
				((char *) pDstVerts)[1] = cY;
				((char *) pDstVerts)[2] = cZ;
				((char *) pDstVerts)[3] = adc;	// Clear ADC bit at 3. vertex
				++pDstVerts;

				// countdown adc
				uADCTimer >>= 1;
				uADCTimerClose >>= 1;

				// Fill up current ST
				const float fTexSize = 128;
				((uint16 *) pDstSTs)[0] = (uint16) (pSTSrcData[0]*fTexSize*16);
				((uint16 *) pDstSTs)[1] = (uint16) (pSTSrcData[1]*fTexSize*16);
				++pDstSTs;

				// Fill up Colors
				((unsigned char *) pDstCols)[0] = (unsigned char) pColSrcData[0];
				((unsigned char *) pDstCols)[1] = (unsigned char) pColSrcData[1];
				((unsigned char *) pDstCols)[2] = (unsigned char) pColSrcData[2];
				((unsigned char *) pDstCols)[3] = (unsigned char) pColSrcData[3];
				++pDstCols;
				
				

				// Calculate next
				if(i<iNumVertsInStrip-1)
				{
					cX = SnapToChar(SnapScale(fRecVx2, fVScale) - SnapScale(fRecVx, fVScale));
					cY = SnapToChar(SnapScale(pVSrcData[4+1], fVScale) - SnapScale(fRecVy, fVScale));
					cZ = SnapToChar(SnapScale(pVSrcData[4+2], fVScale) - SnapScale(fRecVz, fVScale));

					// keep backing up
					fRecVx = fRecVx2, fRecVy = pVSrcData[4+1];
					fRecVz = pVSrcData[4+2], fRecVw = pVSrcData[4+3];
					fRecVx2 = pVSrcData[4+4];
				}

				pVSrcData += 4; pSTSrcData += 4; pColSrcData += 4;
			}
		}
	}

	////////// setup DmaChain ///////////
	/////////////////////////////////////
	
	InitChain(pDmaVif1, STCYCL(1,1));
	AddWordToChain(pDmaVif1, UNPACK(V4_32, 3, PAD_CHUNK+17));

	// add triangle (GifTag)
	AddWordToChain(pDmaVif1, ((uint32 *) orgGIFTagTriangle)[0]);
	AddWordToChain(pDmaVif1, ((uint32 *) orgGIFTagTriangle)[1]);
	AddWordToChain(pDmaVif1, ((uint32 *) orgGIFTagTriangle)[2]);
	AddWordToChain(pDmaVif1, ((uint32 *) orgGIFTagTriangle)[3]);

	// add strip (GifTag)
	AddWordToChain(pDmaVif1, ((uint32 *) orgGIFTagStrip)[0]);
	AddWordToChain(pDmaVif1, ((uint32 *) orgGIFTagStrip)[1]);
	AddWordToChain(pDmaVif1, ((uint32 *) orgGIFTagStrip)[2]);
	AddWordToChain(pDmaVif1, ((uint32 *) orgGIFTagStrip)[3]);

	// deliver vert scale
	float fVs = fMaxDeltaSizeVert / 126.0f;
	AddWordToChain(pDmaVif1, *((uint32 *) &fVs));
	AddWordToChain(pDmaVif1, 0);
	AddWordToChain(pDmaVif1, 0);
	AddWordToChain(pDmaVif1, 0);


	// Create Preload
	uint32 uCurBuf = 0;
	uint32 * pVPtr = (uint32 *) m_pChunks[0].m_pVerts;
	int iStripStart, iNumSkipped;
	AddWordToChain(pDmaVif1, STCYCL(1,1));
	AddWordToChain(pDmaVif1, STMOD(2));
	pVPtr = PreLoad4Verts(pDmaVif1, m_pChunks[0], 0, uCurBuf+19, &iStripStart, &iNumSkipped);
	
	// Insert init regs call
	AddWordToChain(pDmaVif1, ITOP(0));
	AddWordToChain(pDmaVif1, MSCALL(0)); 	

	// add all chunks
	for(int iCount=0; iCount<m_iNumOfChunks; iCount++)
	{
		SStripChunk &Chnk = m_pChunks[iCount];

		assert(Chnk.m_iNumElemsInChunk<=110);

		// toggle buffer
		// room for 110verts, 1 Giftag 
		// and 3 quads of local info
		uCurBuf += (3*110+1+3);
		if(uCurBuf > 2*(3*110+1+3)) uCurBuf = 0;

		// set vif regs
		AddWordToChain(pDmaVif1, STCYCL(1,3));
		AddWordToChain(pDmaVif1, STMOD(2));

		// add vertices
		uint32 uVertsPassed = 0;
		
		for(int iS=iStripStart; iS<Chnk.m_iNumOfStrips; iS++)
		{
			int iNumVertsInCurStrip = Chnk.m_pNumVertsInStrip[iS];

			// if we are recovering from a preload
			// then fix iNumVertsInCurStrip with
			// the preload size, if not then
			// just start a new strip by SetRow
			if(iS==iStripStart) iNumVertsInCurStrip-=iNumSkipped;
			else pVPtr = SetRow(pDmaVif1, pVPtr); 

			uint32 uOffs = PAD_CHUNK+2+3*uVertsPassed+uCurBuf+1;
			AddWordToChain(pDmaVif1, UNPACK(V4_8, iNumVertsInCurStrip, uOffs) );
			
			for(int iN=0; iN<iNumVertsInCurStrip; iN++)
				AddWordToChain(pDmaVif1, *pVPtr++);

			// end
			uVertsPassed += iNumVertsInCurStrip;

			// if we have reached the last strip and
			// if there are more chunks, then preload
			// the next four vertices.
			if(iS==Chnk.m_iNumOfStrips-1 && iCount<m_iNumOfChunks-1)
				pVPtr = PreLoad4Verts(pDmaVif1, m_pChunks[iCount+1], uVertsPassed, uCurBuf, &iStripStart, &iNumSkipped);
			else if(iS==Chnk.m_iNumOfStrips-1 && iCount==m_iNumOfChunks-1)
			{
				// We just finished the last strip in the last chunk
				// now we need to scatter the 4 last ADC values
				uint32 uOffs = PAD_CHUNK+2+3*uVertsPassed+uCurBuf+1;
				AddWordToChain(pDmaVif1, UNPACK(S_8, 4, uOffs) );
				char data[4];
				for(int k=0; k<4; k++)
				{
					if((uADCTimer&1)==1)
					{
						bADC = true;	// set to 0x8000
						assert(uADCTimerClose == 0);
						uADCTimerClose = 0x4;
					}
					else if((uADCTimerClose&1)==1)
					{
						bADC = false;
					}

					char adc = 0;
					if(bADC == true && uRowWfield == 0x7fff)
					{ adc = 1; uRowWfield = 0x8000; }
					else if(bADC == false && uRowWfield == 0x8000)
					{ adc = -1; uRowWfield = 0x7fff; } 

					data[k] = adc;
				
					// countdown adc
					uADCTimer >>= 1;
					uADCTimerClose >>= 1;
				}
				AddWordToChain(pDmaVif1, *((uint32 *) data));
			}
		}
		
		// add TexCrds
		AddWordToChain(pDmaVif1, STMOD(0));
		AddWordToChain(pDmaVif1, UNPACK_UNSIGNED(V2_16, Chnk.m_iNumElemsInChunk, 0+PAD_CHUNK+(uCurBuf+1)) );		
		uint32 * pSTPtr = (uint32 *) Chnk.m_pTexCrds;
		for(int iST=0; iST<Chnk.m_iNumElemsInChunk; iST++)
			AddWordToChain(pDmaVif1, *pSTPtr++);
				
		// Deliver Colours
		AddWordToChain(pDmaVif1, UNPACK_UNSIGNED(V4_8, Chnk.m_iNumElemsInChunk, 1+PAD_CHUNK+(uCurBuf+1)) );
		uint32 * pColPtr = (uint32 *) Chnk.m_pCols;
		for(int iCol=0; iCol<Chnk.m_iNumElemsInChunk; iCol++)
			AddWordToChain(pDmaVif1, *pColPtr++);
			
		
		// Deliver num of strips in current chunk
		AddWordToChain(pDmaVif1, STCYCL(1,1));
		AddWordToChain(pDmaVif1, ITOP(uCurBuf+PAD_CHUNK+1));
		AddWordToChain(pDmaVif1, UNPACK(V4_32, 1, 0+uCurBuf) );
		AddWordToChain(pDmaVif1, (uint32) Chnk.m_iNumElemsInChunk);
		AddWordToChain(pDmaVif1, (uint32) Chnk.m_iNumOfStrips);
		AddWordToChain(pDmaVif1, 0);
		AddWordToChain(pDmaVif1, 0);

		// add draw call
		AddWordToChain(pDmaVif1, MSCALL(2) );
	}

	AddWordToChain(pDmaVif1, FLUSH);
	AddWordToChain(pDmaVif1, FLUSHA);

	// Terminate (add a ret_tag)
	Terminate(pDmaVif1);
	
	return true;
}


void IMesh::CleanUp()
{
	if(m_pChunks!=NULL)
	{
		for(int iCnt=0; iCnt<m_iNumOfChunks; iCnt++)
		{
			if(m_pChunks[iCnt].m_pNumVertsInStrip!=NULL)
				delete [] m_pChunks[iCnt].m_pNumVertsInStrip;

			if(m_pChunks[iCnt].m_pVerts!=NULL)
				delete [] m_pChunks[iCnt].m_pVerts;

			if(m_pChunks[iCnt].m_pCols!=NULL)
				delete [] m_pChunks[iCnt].m_pCols;

			if(m_pChunks[iCnt].m_pTexCrds!=NULL)
				delete [] m_pChunks[iCnt].m_pTexCrds;
		}

		delete [] m_pChunks;
	}

	m_iNumOfChunks = 0;
}

IMesh::IMesh()
{
	m_iNumOfChunks = 0;
	m_pChunks = NULL;
}
