/*
    Copyright (C) 2002 Terratron Technologies Inc.  All Rights Reserved.
    Author: Steven Osman, Morten Mikkelsen, Lionel Lemarie

    This file is part of sps2.

    Before using this file you MUST agree to the license agreement in
    the file LICENSE provided with this package.

    For more information visit:
    http://www.playstation2-linux.com/projects/sps2/
    http://window.terratron.com/~sosman/ps2linux/
    http://www.terratron.com/

    7/12/2003 - Sauce
    Initial version
*/
#include <stdio.h>

#include <sps2lib.h>
#include <sps2tags.h>
#include <sps2util.h>

void drawTriangle();

// Define the coordinates of our triangle
typedef int vertex_t[2];
vertex_t aVertices[]={
  {-150, -150},  // Up from the center, to the left
  { 150, -150},  // Up from the center, to the right
  {   0,  150}   // Down from the center, in the middle
};  // triangle coordinates

// Define the colors of the points
// Make sure there's one of these for each coordinate!
typedef int color_t[3];
color_t aColors[]={
  {255,  0,  0}, // All red
  {  0,255,  0}, // All green
  {  0,  0,255}  // All blue
}; // triangle colors

/**
 * This is the main function of our tutorial.  It will perform some simple
 * initialization and then set up a small loop to repeatedly draw our
 * triangle.
 */
int main() {
  int iSPS2Descriptor;      // The handle we use for sps2

  iSPS2Descriptor=sps2Init();       // Initialize sps2

  // Initialize the screen.  Zero tells sps2UScreenInit to exit gracefully
  // when it receives a signal.
  sps2UScreenInit(0);

  while (1) {
    // Clear the draw area.   The three zeros mean we want to clear it to a
    // color that has zero red, zero green, and zero blue respectively.
    // In simpler terms, we're clearing the screen to black.
    sps2UScreenClear(0,0,0);

    // Draw the triangle
    drawTriangle();

    // Swap displays now that we're done
    sps2UScreenSwap();
  }
  
  // Shut down the screen
  sps2UScreenShutdown();

  // and close sps2
  sps2Release(iSPS2Descriptor);
  return 0;
}

/**
 * This function will draw our triangle.  It's a little bit inefficient because
 * it builds the whole GIF tag every frame (the GIF tag could have been
 * prepared earlier, for instance).  On the other hand, since we're only
 * drawing one triangle on the screen, this is of very little concern.
 */
void drawTriangle() {
  sps2GIFTag_t gifTag;
  sps2GIFPackedRegister_t gifRegister;
  int iPointLooper;

  // Prepare the GIF tag  
  gifTag.i128=0;       // Blank it out
  gifTag.s.NLOOP=3;    // 3 entries.  One entry for each of the points.
  gifTag.s.PRE=1;      // We are providing a valid PRIM value
  gifTag.s.EOP=1;      // End of packet -- draw me please!
  gifTag.s.PRIM=11;    // 3 = triangle | 8 = gourard shading
  gifTag.s.FLG=GIF_FLG_PACKED; // We're using the packed format for data
  gifTag.s.NREG=0;     // Start off with zero registers, increment as we
                         // populate the descriptions
  
  // Add a register.  It will be an RGBAQ register
  SPS2_SET_GIF_REG(gifTag, gifTag.s.NREG++, GIF_REG_RGBAQ);
  // Add another register.  It will by an XYZ2 register
  SPS2_SET_GIF_REG(gifTag, gifTag.s.NREG++, GIF_REG_XYZ2);

  // Send the GIF tag
  DPUT_EE_GIF_FIFO(gifTag.i128);

  // Now that we promised three sets of RGBAQ and XYZ2 registers, we should
  // deliver them.
  for (iPointLooper=0; iPointLooper<3;iPointLooper++) {
    // First, prepare and sent the RGBAQ register
    gifRegister.i128=0;
    gifRegister.RGBAQ.s.R=aColors[iPointLooper][0];
    gifRegister.RGBAQ.s.G=aColors[iPointLooper][1];
    gifRegister.RGBAQ.s.B=aColors[iPointLooper][2];
    DPUT_EE_GIF_FIFO(gifRegister.i128);

    // Next, prepare and send the XYZ2 register
    gifRegister.i128=0;
    gifRegister.XYZ2.s.X=(aVertices[iPointLooper][0]+2048) << 4;
    gifRegister.XYZ2.s.Y=(aVertices[iPointLooper][1]+2048) << 4;
    gifRegister.XYZ2.s.Z=128; // Z has no decimal places, it needs no << 4
    DPUT_EE_GIF_FIFO(gifRegister.i128);
  }
}
